from kilroy_module_huggingface.modules.base import HuggingfaceModule
from kilroy_module_huggingface.modules.basic import BasicHuggingfaceModule
from kilroy_module_huggingface.modules.reward import (
    RewardModelHuggingfaceModule,
)



import re
from abc import ABC
from typing import Optional

import torch
from kilroy_module_pytorch_py_sdk import (
    LanguageModel,
    RewardModel,
    pack_padded,
    unpack_to_padded,
)
from torch import Tensor, nn
from torch.nn.utils.rnn import PackedSequence
from transformers import (
    AutoConfig,
    AutoModelForCausalLM,
    AutoModelForSequenceClassification,
    PreTrainedModel,
)


def make_mask(x: Tensor, lengths: Tensor) -> Tensor:
    indices = torch.arange(x.shape[-1]).repeat((len(x), 1))
    lengths = lengths.view(-1, 1)
    return indices < lengths


class HuggingfaceModelBase(ABC):
    def __init__(
        self,
        model: PreTrainedModel,
        pad_token_id: int,
    ) -> None:
        super().__init__()
        self._model = model
        self._pad_token_id = pad_token_id

    @property
    def base_model(self) -> nn.Module:
        return self._model.base_model

    def freeze(self, pattern: Optional[str] = ".*") -> None:
        if pattern is not None:
            pattern = re.compile(pattern)

        for name, parameter in self._model.base_model.named_parameters():
            if pattern is not None and pattern.match(name):
                parameter.requires_grad = False


class HuggingfaceLanguageModel(HuggingfaceModelBase, LanguageModel):
    @classmethod
    def from_path(cls, path: str) -> "HuggingfaceLanguageModel":
        model = AutoModelForCausalLM.from_pretrained(path)
        if model.config.pad_token_id is not None:
            pad_token_id = model.config.pad_token_id
        elif model.config.eos_token_id is not None:
            pad_token_id = model.config.eos_token_id
        else:
            pad_token_id = 0
        return cls(model, pad_token_id)

    def freeze(self, pattern: Optional[str] = ".*") -> None:
        super().freeze(pattern)

        for parameter in self._model.lm_head.parameters():
            parameter.requires_grad = True

    def forward(self, x: PackedSequence) -> PackedSequence:
        x, lengths = unpack_to_padded(x, pad_value=self._pad_token_id)
        x = x[:, :, 0]
        mask = make_mask(x, lengths)
        y = self._model(x, attention_mask=mask)
        return pack_padded(y.logits.log_softmax(-1), lengths)


class HuggingfaceRegressionModel(HuggingfaceModelBase, RewardModel):
    @classmethod
    def from_path(cls, path: str) -> "HuggingfaceRegressionModel":
        config = AutoConfig.from_pretrained(path)
        if config.pad_token_id is not None:
            pad_token_id = config.pad_token_id
        elif config.eos_token_id is not None:
            pad_token_id = config.eos_token_id
        else:
            pad_token_id = 0
        model = AutoModelForSequenceClassification.from_pretrained(
            path,
            num_labels=1,
            problem_type="regression",
            pad_token_id=pad_token_id,
        )
        return cls(model, pad_token_id)

    def freeze(self, pattern: Optional[str] = ".*") -> None:
        super().freeze(pattern)

        for parameter in self._model.score.parameters():
            parameter.requires_grad = True

    def forward(self, x: PackedSequence) -> Tensor:
        x, lengths = unpack_to_padded(x, pad_value=self._pad_token_id)
        x = x[:, :, 0]
        mask = make_mask(x, lengths)
        return self._model(x, attention_mask=mask).logits



from typing import List

from kilroy_module_pytorch_py_sdk import Tokenizer
from transformers import AutoTokenizer, PreTrainedTokenizerBase


class HuggingfaceTokenizer(Tokenizer):
    def __init__(self, tokenizer: PreTrainedTokenizerBase) -> None:
        super().__init__()
        self._tokenizer = tokenizer

    @classmethod
    def from_path(cls, path: str) -> "HuggingfaceTokenizer":
        return cls(AutoTokenizer.from_pretrained(path))

    def encode(self, text: str) -> List[int]:
        indices = self._tokenizer.encode(text)

        if len(indices) <= 1:
            indices = (
                [self._tokenizer.bos_token_id]
                + indices
                + [self._tokenizer.eos_token_id]
            )

        if indices[0] != self._tokenizer.bos_token_id:
            indices = [self._tokenizer.bos_token_id] + indices
        if indices[-1] != self._tokenizer.eos_token_id:
            indices = indices + [self._tokenizer.eos_token_id]

        return indices

    def decode(self, indices: List[int]) -> str:
        return self._tokenizer.decode(indices, skip_special_tokens=True)

    @property
    def start_token(self) -> int:
        return self._tokenizer.bos_token_id

    @property
    def end_token(self) -> int:
        return self._tokenizer.eos_token_id



import asyncio
from asyncio import Queue
from typing import Any, Coroutine, Dict, Optional

from kilroy_module_pytorch_py_sdk import (
    Codec,
    Configurable,
    Generator,
    Metadata,
    Optimizer,
    RewardModelModule,
    RewardModelModuleState,
    SerializableModel,
    background,
    classproperty,
)
from kilroy_module_pytorch_py_sdk.modules.reward import (
    ReinforcedScoreMetric,
    RewardModelLossMetric,
    RewardModelScoreMetric,
    SupervisedLossMetric,
)

from kilroy_module_huggingface.models import (
    HuggingfaceLanguageModel,
    HuggingfaceRegressionModel,
)
from kilroy_module_huggingface.modules.base import HuggingfaceModule
from kilroy_module_huggingface.tokenizer import HuggingfaceTokenizer


class ModelParams(SerializableModel):
    name: str
    freeze: Optional[str] = None
    optimizer_type: str = "adam"
    optimizers_params: Dict[str, Dict[str, Any]] = {}


class Params(SerializableModel):
    language_model_params: ModelParams
    reward_model_params: ModelParams
    frontend_generator_params: Dict[str, Any] = {}
    backend_generator_params: Dict[str, Any] = {}
    codec_params: Dict[str, Any] = {}
    batch_size: int
    sample_size: int


class RewardModelHuggingfaceModule(
    RewardModelModule, HuggingfaceModule[RewardModelModuleState]
):
    @classproperty
    def metadata(cls) -> Metadata:
        return Metadata(
            key="kilroy-module-huggingface",
            description="Kilroy module for Huggingface models",
        )

    @staticmethod
    async def _build_language_model(
        params: Params,
    ) -> HuggingfaceLanguageModel:
        return await background(
            HuggingfaceLanguageModel.from_path,
            params.language_model_params.name,
        )

    @staticmethod
    async def _build_reward_model(
        params: Params,
    ) -> HuggingfaceRegressionModel:
        return await background(
            HuggingfaceRegressionModel.from_path,
            params.reward_model_params.name,
        )

    @staticmethod
    async def _build_language_model_tokenizer(
        params: Params,
    ) -> HuggingfaceTokenizer:
        return await background(
            HuggingfaceTokenizer.from_path, params.language_model_params.name
        )

    @staticmethod
    async def _build_reward_model_tokenizer(
        params: Params,
    ) -> HuggingfaceTokenizer:
        return await background(
            HuggingfaceTokenizer.from_path, params.reward_model_params.name
        )

    @staticmethod
    async def _build_language_model_optimizer(
        params: Params, model: HuggingfaceLanguageModel
    ) -> Optimizer:
        opt_cls = Optimizer.for_category(
            params.language_model_params.optimizer_type
        )
        opt_params = params.language_model_params.optimizers_params.get(
            params.language_model_params.optimizer_type, {}
        )
        if issubclass(opt_cls, Configurable):
            optimizer = await opt_cls.build(
                params=model.parameters(), **opt_params
            )
            await optimizer.init()
        else:
            optimizer = opt_cls(model.parameters(), **opt_params)
        return optimizer

    @staticmethod
    async def _build_reward_model_optimizer(
        params: Params, model: HuggingfaceRegressionModel
    ) -> Optimizer:
        opt_cls = Optimizer.for_category(
            params.reward_model_params.optimizer_type
        )
        opt_params = params.reward_model_params.optimizers_params.get(
            params.reward_model_params.optimizer_type, {}
        )
        if issubclass(opt_cls, Configurable):
            optimizer = await opt_cls.build(
                params=model.parameters(), **opt_params
            )
            await optimizer.init()
        else:
            optimizer = opt_cls(model.parameters(), **opt_params)
        return optimizer

    @staticmethod
    async def _build_frontend_generator(params: Params) -> Generator:
        generator = await Generator.build(**params.frontend_generator_params)
        await generator.init()
        return generator

    @staticmethod
    async def _build_backend_generator(params: Params) -> Generator:
        generator = await Generator.build(**params.backend_generator_params)
        await generator.init()
        return generator

    @staticmethod
    async def _build_codec(params: Params) -> Codec:
        codec = await Codec.build(**params.codec_params)
        await codec.init()
        return codec

    async def build_default_state(self) -> RewardModelModuleState:
        params = Params(**self._kwargs)
        language_model = await self._build_language_model(params)
        reward_model = await self._build_reward_model(params)
        language_model_optimizer = await self._build_language_model_optimizer(
            params, language_model
        )
        reward_model_optimizer = await self._build_reward_model_optimizer(
            params, reward_model
        )
        language_model.freeze(params.language_model_params.freeze)
        reward_model.freeze(params.reward_model_params.freeze)
        coroutine_queue = Queue()
        return RewardModelModuleState(
            language_model=language_model,
            reward_model=reward_model,
            language_model_tokenizer=await self._build_language_model_tokenizer(
                params
            ),
            reward_model_tokenizer=await self._build_reward_model_tokenizer(
                params
            ),
            language_model_optimizer=language_model_optimizer,
            language_model_optimizers_params=params.language_model_params.optimizers_params,
            reward_model_optimizer=reward_model_optimizer,
            reward_model_optimizers_params=params.reward_model_params.optimizers_params,
            frontend_generator=await self._build_frontend_generator(params),
            backend_generator=await self._build_backend_generator(params),
            codec=await self._build_codec(params),
            results_cache={},
            batch_size=params.batch_size,
            sample_size=params.sample_size,
            epoch=0,
            supervised_loss_metric=await SupervisedLossMetric.build(),
            reinforced_score_metric=await ReinforcedScoreMetric.build(),
            reward_model_loss_metric=await RewardModelLossMetric.build(),
            reward_model_score_metric=await RewardModelScoreMetric.build(),
            epoch_supervised_losses=[],
            epoch_reinforced_scores=[],
            epoch_reward_model_losses=[],
            epoch_reward_model_scores=[],
            coroutine_queue=coroutine_queue,
            worker_task=asyncio.create_task(self._work(coroutine_queue)),
        )

    @staticmethod
    async def _work(queue: Queue[Coroutine]) -> None:
        while True:
            coroutine = await queue.get()
            await coroutine



"""Main script.

This module provides basic CLI entrypoint.

"""
import asyncio
import logging
from asyncio import FIRST_EXCEPTION
from enum import Enum
from logging import Logger
from typing import Dict, Optional

import typer
from kilroy_module_pytorch_py_sdk import ModuleServer
from typer import FileText

from kilroy_module_huggingface.config import get_config
from kilroy_module_huggingface.modules import HuggingfaceModule

cli = typer.Typer()  # this is actually callable and thus can be an entry point


class Verbosity(str, Enum):
    DEBUG = "DEBUG"
    INFO = "INFO"
    WARNING = "WARNING"
    ERROR = "ERROR"
    CRITICAL = "CRITICAL"


def get_logger(verbosity: Verbosity) -> Logger:
    logging.basicConfig()
    logger = logging.getLogger("kilroy-module-huggingface")
    logger.setLevel(verbosity.value)
    return logger


async def run(config: Dict, logger: Logger) -> None:
    module_cls = HuggingfaceModule.for_category(config["moduleType"])
    module = await module_cls.build(**config.get("moduleParams", {}))

    server = ModuleServer(module, logger)

    tasks = (
        asyncio.create_task(module.init()),
        asyncio.create_task(server.run(**config.get("serverParams", {}))),
    )

    try:
        done, pending = await asyncio.wait(tasks, return_when=FIRST_EXCEPTION)
    except asyncio.CancelledError:
        done, pending = [], tasks

    for task in pending:
        task.cancel()
        try:
            await task
        except asyncio.CancelledError:
            pass

    for task in done:
        task.result()

    await module.cleanup()


@cli.command()
def main(
    config: Optional[FileText] = typer.Option(
        default=None, help="Configuration file"
    ),
    verbosity: Verbosity = typer.Option(
        default="INFO", help="Verbosity level."
    ),
) -> None:
    """Command line interface for kilroy-module-huggingface."""

    config = get_config(config)
    logger = get_logger(verbosity)

    asyncio.run(run(config, logger))


if __name__ == "__main__":
    # entry point for "python -m"
    cli()



from copy import deepcopy
from typing import Any, Dict, Optional, TextIO

import yaml
from deepmerge import Merger

from kilroy_module_huggingface import resource_text

_DEFAULT_MERGER = Merger(
    [(list, ["override"]), (dict, ["merge"]), (set, ["override"])],
    ["override"],
    ["override"],
)


def _merge_configs(
    default_config: Dict[str, Any],
    user_config: Dict[str, any],
    merger: Merger = _DEFAULT_MERGER,
    **kwargs,
) -> Dict[str, Any]:
    config = deepcopy(default_config)
    merger.merge(config, user_config)
    merger.merge(config, kwargs)
    return config


def get_config(f: Optional[TextIO] = None, **kwargs) -> Dict[str, Any]:
    config = yaml.safe_load(resource_text("config.yaml"))
    user_config = yaml.safe_load(f) if f else {}
    return _merge_configs(config, user_config, **kwargs)



from typing import Any, Dict, Optional

from kilroy_module_pytorch_py_sdk import (
    BasicModule,
    BasicModuleState,
    Codec,
    Configurable,
    Generator,
    Metadata,
    Optimizer,
    SerializableModel,
    background,
    classproperty,
)
from kilroy_module_pytorch_py_sdk.modules.basic import (
    ReinforcedScoreMetric,
    SupervisedLossMetric,
)

from kilroy_module_huggingface.models import HuggingfaceLanguageModel
from kilroy_module_huggingface.modules.base import HuggingfaceModule
from kilroy_module_huggingface.tokenizer import HuggingfaceTokenizer


class Params(SerializableModel):
    model_name: str
    freeze: Optional[str] = None
    optimizer_type: str = "adam"
    optimizers_params: Dict[str, Dict[str, Any]] = {}
    generator_params: Dict[str, Any] = {}
    codec_params: Dict[str, Any] = {}
    batch_size: int


class BasicHuggingfaceModule(BasicModule, HuggingfaceModule[BasicModuleState]):
    @classproperty
    def metadata(cls) -> Metadata:
        return Metadata(
            key="kilroy-module-huggingface",
            description="Kilroy module for Huggingface models",
        )

    @staticmethod
    async def _build_model(
        params: Params,
    ) -> HuggingfaceLanguageModel:
        return await background(
            HuggingfaceLanguageModel.from_path, params.model_name
        )

    @staticmethod
    async def _build_tokenizer(
        params: Params,
    ) -> HuggingfaceTokenizer:
        return await background(
            HuggingfaceTokenizer.from_path, params.model_name
        )

    @staticmethod
    async def _build_optimizer(
        params: Params, model: HuggingfaceLanguageModel
    ) -> Optimizer:
        opt_cls = Optimizer.for_category(params.optimizer_type)
        opt_params = params.optimizers_params.get(params.optimizer_type, {})
        if issubclass(opt_cls, Configurable):
            optimizer = await opt_cls.build(
                params=model.parameters(), **opt_params
            )
            await optimizer.init()
        else:
            optimizer = opt_cls(model.parameters(), **opt_params)
        return optimizer

    @staticmethod
    async def _build_generator(params: Params) -> Generator:
        generator = await Generator.build(**params.generator_params)
        await generator.init()
        return generator

    @staticmethod
    async def _build_codec(params: Params) -> Codec:
        codec = await Codec.build(**params.codec_params)
        await codec.init()
        return codec

    async def build_default_state(self) -> BasicModuleState:
        params = Params(**self._kwargs)
        model = await self._build_model(params)
        optimizer = await self._build_optimizer(params, model)
        model.freeze(params.freeze)
        return BasicModuleState(
            model=model,
            tokenizer=await self._build_tokenizer(params),
            optimizer=optimizer,
            optimizers_params=params.optimizers_params,
            generator=await self._build_generator(params),
            codec=await self._build_codec(params),
            results_cache={},
            batch_size=params.batch_size,
            epoch=0,
            supervised_loss_metric=await SupervisedLossMetric.build(),
            reinforced_score_metric=await ReinforcedScoreMetric.build(),
            epoch_supervised_losses=[],
            epoch_reinforced_scores=[],
        )



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['kilroy_module_huggingface',
 'kilroy_module_huggingface.modules',
 'kilroy_module_huggingface.resources']

package_data = \
{'': ['*']}

install_requires = \
['deepmerge>=1.0,<2.0',
 'kilroy-module-pytorch-py-sdk>=0.4,<0.5',
 'transformers>=4.21,<5.0',
 'typer[all]>=0.4,<0.5']

extras_require = \
{'dev': ['pytest>=7.0,<8.0'], 'test': ['pytest>=7.0,<8.0']}

entry_points = \
{'console_scripts': ['kilroy-module-huggingface = '
                     'kilroy_module_huggingface.__main__:cli']}

setup_kwargs = {
    'name': 'kilroy-module-huggingface',
    'version': '0.4.0',
    'description': 'kilroy module using Hugging Face models 🤗',
    'long_description': '<h1 align="center">kilroy-module-huggingface</h1>\n\n<div align="center">\n\nkilroy module using Hugging Face models 🤗\n\n[![Multiplatform tests](https://github.com/kilroybot/kilroy-module-huggingface/actions/workflows/test-multiplatform.yml/badge.svg)](https://github.com/kilroybot/kilroy-module-huggingface/actions/workflows/test-multiplatform.yml)\n[![Docker tests](https://github.com/kilroybot/kilroy-module-huggingface/actions/workflows/test-docker.yml/badge.svg)](https://github.com/kilroybot/kilroy-module-huggingface/actions/workflows/test-docker.yml)\n[![Docs](https://github.com/kilroybot/kilroy-module-huggingface/actions/workflows/docs.yml/badge.svg)](https://github.com/kilroybot/kilroy-module-huggingface/actions/workflows/docs.yml)\n\n</div>\n\n---\n\n## Installing\n\nUsing `pip`:\n\n```sh\npip install kilroy-module-huggingface\n```\n',
    'author': 'kilroy',
    'author_email': 'kilroymail@pm.me',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/kilroybot/kilroy-module-huggingface',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from abc import ABC
from typing import Generic, TypeVar

from kilroy_module_server_py_sdk import (
    Categorizable,
    Module,
    classproperty,
    normalize,
)

StateType = TypeVar("StateType")


class HuggingfaceModule(
    Categorizable, Module[StateType], Generic[StateType], ABC
):
    @classproperty
    def category(cls) -> str:
        name: str = cls.__name__
        return normalize(name.removesuffix("HuggingfaceModule"))
